import torch
from random import shuffle
import numpy as np

## --- Default Architectures ---
# `**hp` must be included

# `select_fn_lose()`
def lose_binary(**hp):
	loser = torch.nn.BCELoss()
	return loser

def lose_multiclass(**hp):
	# ptrckblck says `nn.NLLLoss()` will work too.
	loser = torch.nn.CrossEntropyLoss()
	return loser

def lose_regression(**hp):
	loser = torch.nn.L1Loss()#mean absolute error.
	return loser

# `select_fn_optimize()`
def optimize(model, **hp):
	optimizer = torch.optim.Adamax(model.parameters(),lr=0.01)
	return optimizer

# `select_fn_predict()`
def predict_binary(model, samples_predict):
	probability = model(samples_predict['features'])
	# Convert tensor back to numpy for AIQC metrics.
	probability = probability.detach().numpy()
	prediction = (probability > 0.5).astype("int32")
	# Both objects are numpy.
	return prediction, probability

def predict_multiclass(model, samples_predict):
	probabilities = model(samples_predict['features'])
	# Convert tensor back to numpy for AIQC metrics.
	probabilities = probabilities.detach().numpy()
	prediction = np.argmax(probabilities, axis=-1)
	# Both objects are numpy.
	return prediction, probabilities

def predict_regression(model, samples_predict):
	prediction = model(samples_predict['features']).detach().numpy()
	return prediction


## --- Batching and Training Loop ---
def drop_invalidBatch(
	batched_data:object
	, batch_size:int
	, enforce_sameSize:bool=True
	, allow_singleSample:bool=False
):
	"""
	`enforce_sameSize=True` Primarily because it influence batch size, therefore layer dimensions, 
	and also because true tensors must have uniform shape.
	"""
	# Similar to a % remainder, this will only apply to the last element in the batch.
	last_batch_size = batched_data[-1].shape[0]
	# If there is a problem, then just trim the last split.
	if (last_batch_size==1):
		if (allow_singleSample==True):
			print("\nWarning - The size of the last batch is 1,\n which commonly leads to PyTorch errors.\nTry using `torch_batch_samples(allow_singleSample=False)\n")
		elif (allow_singleSample==False): 
			batched_data = batched_data[:-1]
	elif ((enforce_sameSize==True) and (batch_size!=last_batch_size)):
		batched_data = batched_data[:-1]
	return batched_data


def shuffle_samples(features:list, labels:list):
	"""Assumes that the first index represents the batch."""
	rand_idx = list(range(len(labels)))
	shuffle(rand_idx)

	features = torch.stack([features[i] for i in rand_idx]).to(torch.float)
	labels = torch.stack([labels[i] for i in rand_idx]).to(torch.float)
	return features, labels


def batch_samples(
	features:object, labels:object
	, batch_size = 5, enforce_sameSize:bool=False, allow_singleSample:bool=False
):
	if (batch_size==1):
		if (allow_singleSample==False):
			raise Exception("\nYikes - `batch_size==1` but `allow_singleSample==False`.\n")
		elif (allow_singleSample==True):
			print("\nWarning - PyTorch errors are common when `batch_size==1`.")
	
	# split() normally returns a tuple.
	features = list(torch.split(features, batch_size))
	labels = list(torch.split(labels, batch_size))

	features = drop_invalidBatch(features, batch_size, enforce_sameSize, allow_singleSample)
	labels = drop_invalidBatch(labels, batch_size, enforce_sameSize, allow_singleSample)
	return features, labels


def shuffle_batches(features:list, labels:list):
	"""
	- Assumes that the first index represents the batch.
	- Makes sure batches aren't seen in same order every epoch.
	"""
	rand_idx = list(range(len(labels)))
	shuffle(rand_idx)
	features = [features[i] for i in rand_idx]
	labels = [labels[i] for i in rand_idx]
	return features, labels


def flatten_uniColumn(tzr:object):
	if (tzr.shape[-1]==1):
		return tzr.flatten()
	return tzr


def float_to_int(tzr:object):
	"""Handles float/int incosistencies of torch's loss and torchmetrics' scoring."""
	if (tzr.type()=='torch.FloatTensor'):
		if all([float(i).is_integer() for i in tzr[:3]]):
			return tzr.to(torch.int64)
			# ^ Sample to see if floats are actually categorical ints `float(0).is_integer()==True`
		else:
			raise Exception(f"\nYikes - Scoring failed on {tzr.type()}.\nDid not attempt as int64 because tensor contained non-zero decimals.\n")
	else:
		raise Exception(f"\nYikes - Scoring failed because {tzr.type()} type not supported.\n")


def fit(
	model:object, loser:object, optimizer:object,  
	samples_train:dict, samples_evaluate:dict,
	epochs:int=30, batch_size:int=5, enforce_sameSize=True, allow_singleSample=False,  
	metrics:list=None
):
	"""
	- This is the only user-facing function for non-manual training loops.
	- It is designed to handle all supervised scenarios.
	- Have not tested this with self-supervised where 2D+ compared to 2D+
	"""
	# Mirrors `tf.keras.model.History.history` schema for use with `Predictor.plot_learning_curve()`
	history = dict(loss=list(), val_loss=list())
	metrics_keys = []
	if (metrics is not None):
		for m in metrics:
			# An initialized metric actually contains `None` so `utils.listify` doesn't work here.
			if ('torchmetrics' not in str(type(m))):
				raise Exception("\nYikes - Did you forget to initialize your metric?\ne.g. do `torchmetrics.Accuracy()`, not `torchmetrics.Accuracy`\n")
			name = m.__class__.__name__
			history[name] = list()
			val_name = f"val_{name}"
			history[val_name] = list()
			metrics_keys.append((name, val_name))
	
	shuffled_features, shuffled_labels = shuffle_samples(samples_train['features'], samples_train['labels'])
	## --- Prepare mini batches for analysis ---
	batched_features, batched_labels = batch_samples(
		shuffled_features, shuffled_labels,
		batch_size=batch_size, enforce_sameSize=enforce_sameSize, allow_singleSample=allow_singleSample
	)
	"""
	- On one hand, I could pass in `analysis_type` to deterministically handle the proper
	  dimensionality and type of the data for loss and metrics.
	- However, performance-wise, that would still require a lot of if statements. 
	- The `try` approach is more future-proof.
	- `flatten()` works on any dimension, even 1D.
	- Remember, multi-label  PyTorch uses ordinal labels, but OHE output probabilities.
	  It wants to compare 2D probabilities to 1D ordinal labels.
	- Unsupervised analysis either succeeds as 2D+ or fails. MSE works on 3D data, but r2 fails. 
	  We could stack 3D+ into 2D, but then we'd have to stack features as well, and that's kind 
	  of insane because this is just for epoch-level metrics.
	"""
	## --- Training Loop ---
	for epoch in range(epochs):
		batched_features, batched_labels = shuffle_batches(batched_features, batched_labels)
		## --- Batch Training ---
		for i, batch in enumerate(batched_features):
			# Make raw (unlabeled) predictions.
			batch_probability = model(batched_features[i])
			batch_labels = batched_labels[i]
			try:
				batch_loss = loser(batch_probability, batch_labels)
			except:
				# Known exception: multi classify fails on 2D and floats
				batch_labels = flatten_uniColumn(batch_labels)
				batch_labels = float_to_int(batch_labels)
				batch_loss = loser(batch_probability, batch_labels)
			# Backpropagation.
			optimizer.zero_grad()
			batch_loss.backward()
			optimizer.step()

		## --- Epoch Loss ---
		# Known exception: multi classify fails on floats
		train_probability = model(samples_train['features'])
		train_probability = flatten_uniColumn(train_probability)
		train_labels = flatten_uniColumn(samples_train['labels'])
		try:
			train_loss = loser(train_probability, train_labels)
		except:
			train_labels = float_to_int(train_labels)
			train_loss = loser(train_probability, train_labels)
		history['loss'].append(float(train_loss))

		eval_probability = model(samples_evaluate['features'])
		eval_probability = flatten_uniColumn(eval_probability)
		eval_labels = flatten_uniColumn(samples_evaluate['labels'])
		try:
			eval_loss = loser(eval_probability, eval_labels)
		except:
			eval_labels = float_to_int(eval_labels)
			eval_loss = loser(eval_probability, eval_labels)
		history['val_loss'].append(float(eval_loss))

		## --- Epoch Metrics ---
		# Known exception: binary classify accuracy fails on floats.
		for i, m in enumerate(metrics):
			try:
				train_m = m(train_probability, train_labels)
			except:
				train_labels = float_to_int(train_labels)
				train_m = m(train_probability, train_labels)
			metrics_key = metrics_keys[i][0]
			history[metrics_key].append(float(train_m))

			try:
				eval_m = m(eval_probability, eval_labels)
			except:
				eval_labels = float_to_int(eval_labels)
				eval_m = m(eval_probability, eval_labels)
			metrics_key = metrics_keys[i][1]
			history[metrics_key].append(float(eval_m))
	return model, history



"""
Serialization
└── Documentation = https://dill.readthedocs.io/en/latest/index.html

Goes beyond pickle to handle nested functions and classes.
Dill-serialized objects are pickleable.
"""
from io import BytesIO
from textwrap import dedent
import dill as dill


def serialize(objekt:object):
	blob = BytesIO()
	dill.dump(objekt, blob)
	blob = blob.getvalue()
	return blob


def deserialize(blob:bytes):
	objekt = BytesIO(blob)
	objekt = dill.load(objekt)
	return objekt


def reveal_code(blob:object, print_it:bool=True):
	code_str = (
		dill.source.getsource(
			deserialize(blob).__code__
		)
	)
	if (print_it == True):
		print(dedent(code_str))
	return code_str



from . import tf_bin_img, tf_bin_seq, tf_bin_tab, tf_bin_text, \
    tf_fore_img, tf_fore_tab, tf_multi_tab, tf_reg_tab, torch_bin_img, \
    torch_bin_tab, torch_multi_tab, torch_reg_tab

"""Register them all here so that they can be called like so: `from aiqc import tests`"""



# Local modules.
from .. import orm
from ..utils.meter import metrics_classify, metrics_regress
# Python modules.
import io
from logging import getLogger, ERROR
from contextlib import redirect_stdout

from dash import dcc, html
from dash.dependencies import Output, Input, State
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from jupyter_dash import JupyterDash
"""
- Using `jupyter_dash` because it runs on a background proc. 
- Which doesn't actually require jupyter; can be run from shell.
- But only `external` mode works in shell and ipython.

Looking forward:
- Docs regarding __name__ and __main__ dash.plotly.com/external-resources
- Serve using Flask stackoverflow.com/a/60026683/5739514
"""
# Bug: <github.com/plotly/jupyter-dash/issues/63>
getLogger('werkzeug').setLevel(ERROR)


class Tracker(object):        
    def __init__(self, refresh_seconds:int=5, server_runtime:dict=None):
        """
            - `refresh_seconds` determines the polling frequency.
            - `server_runtime` is passed through to `dash.Dash.app.run_server()` as **kwargs.
        """
        if ((refresh_seconds < 5) or (refresh_seconds > 30)):
            raise Exception("\nYikes - `refresh_rate` must be between 5 and 30 seconds.\n")
        refresh_seconds *= 1000

        runtime = dict(
            # `external` browser tab | `inline` jupyter cell | `jupyterlab` jupyter tab
            mode = "external"
            # Examining errors and callback DAG.
            , debug = False 
            , host = '127.0.0.1' 
            # Just keep incrementing if it's "already in use".
            , port = '9991'
        )
        if (server_runtime is not None):
            for parameter, value in server_runtime.items():
                runtime[parameter] = value
        
        self.refresh_seconds = refresh_seconds
        self.server_runtime = runtime
        self._clean_run = None


    def stop(self):
        if (self._clean_run is None):
            raise Exception("\nYikes - Cannot stop app because it is not running yet.\n")
        elif (self._clean_run is False):
            raise Exception("\nYikes - Cannot stop app due to known bug:\nhttps://github.com/plotly/jupyter-dash/issues/33\n")
        else:
            JupyterDash._terminate_server_for_port(
                self.server_runtime['host'], self.server_runtime['port'],
            )
            print("🚥 AIQC Tracker stopped 🚥")


    def start(self):
        """
        - "External css/js files are loaded before the `/assets`."
        - "CSS/JS files in `/assets` are auto served." Don't need to be added as externals.
        - BOOTSTRAP points to remote cdn: https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css
        """
        sheets = [dbc.themes.BOOTSTRAP]
        app = JupyterDash(__name__, external_stylesheets=sheets, update_title = "♻️")
        # Mandatory when using components that are generated by other callbacks
        app.config['suppress_callback_exceptions'] = True
        # `app.update_title` can only be set when instantiating the app.
        app.title = "AIQC Lab"


        app.layout = html.Div(
            [
                dcc.Interval(id="initial_load", n_intervals=0, max_intervals=-1, interval=self.refresh_seconds),
                # ====== NAVBAR ======
                html.Div(
                    html.Center(
                        html.A(
                            html.Img(src="assets/logo_wide_small.svg", height="35px"),
                            href="https://aiqc.readthedocs.io",
                        ),
                        className='logo'
                    ),
                    className='navig'
                ),
                # Grid system Div(Row(Col(Div)))
                # dash-bootstrap-components.opensource.faculty.ai/docs/components/layout/
                # ====== EXPERIMENT ======
                dbc.Row(
                    [
                        # =CONTROLS=
                        dbc.Col(
                            dbc.InputGroup(
                                [
                                    dbc.InputGroupText("Experiment ID"),
                                    # Not a callback because it is the first inputable object.
                                    dbc.Select(id='exp_dropdown'),
                                ],
                                size="sm", className='ctrl_chart ctrl_big ctr'
                            ),
                            width=3, align="center", className='ctrl_col_big'
                        ),
                        dbc.Col(
                            dbc.InputGroup(
                                [
                                    dbc.InputGroupText("Score Type"),
                                    # Defined by downstream callback
                                    dbc.Select(id='exp_type'),
                                ],
                                size="sm", className='ctrl_chart ctrl_big ctr'
                            ),
                            width=3, align="center", className='ctrl_col_big'
                        ),
                        dbc.Col(
                            dbc.InputGroup(
                                [
                                    dbc.InputGroupText("Min Score"),
                                    dbc.Input(
                                        type="number",id='exp_score',
                                        placeholder="-∞"
                                    ),
                                ],
                                size="sm", className='ctrl_chart ctrl_sm ctr'
                            ),
                            width=2, align="center", className='ctrl_col_sm'
                        ),
                        dbc.Col(
                            dbc.InputGroup(
                                [
                                    dbc.InputGroupText("Max Loss"),
                                    dbc.Input(
                                        type="number",id='exp_loss',
                                        placeholder="+∞"
                                    ),
                                ],
                                size="sm", className='ctrl_chart ctrl_sm ctr'
                            ),
                            width=2, align="center", className='ctrl_col_sm'
                        ),
                        dbc.Col(
                            dbc.InputGroup(
                                [
                                    dbc.Button(
                                        "Filter", outline=True, 
                                        n_clicks=0, id="exp_button",
                                        className='chart_button ctr',
                                    ),
                                ],
                                size="sm", className='ctrl_chart ctr'
                            ),
                            width=1, align="center",
                        ),
                    ],
                    className='exp_ctrl_row'
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            html.Div(
                                dbc.Spinner(
                                    color="info", delay_hide=1,
                                    spinner_style={"width": "4rem", "height": "4rem"}
                                ),
                                className='spinner'
                            ),
                            id="exp_plot_contain", width=8, align="center"
                        ),
                        dbc.Col(
                            dbc.Alert(
                                "Click on a model in the chart on the left to show its parameters.",
                                className='alert'
                            ),
                            id="param_pane", className='hp_contain',
                            width=4, align="center"
                        )
                    ],
                    className='exp_row'
                ),
                # ====== MIDDLE BAR ======
                # Stacked so sticks to top regardless of progress bar presence 
                html.Div(
                    [
                        # =PROGRESS BAR=
                        # w/o gutter class the rounded edges of bar don't fill.
                        dbc.Row(id="progress_container", className='g-0'),
                        # =OPTIONS=
                        dbc.Row(
                            [
                                dbc.Col(width="4"),
                                dbc.Col(
                                    dcc.Dropdown(id="multitron", multi=True, className='multitron ctr'),
                                    width="4", align="center",
                                ),
                                dbc.Col(width="3"),
                                dbc.Col(
                                    html.Div(html.A("⇪",href="#",className='up_arrow')),
                                    width="1", align="center",
                                )
                            ]
                        )
                    ],
                    className='middle_bar'
                ),
                html.Br(),html.Br(),html.Br(),
                dbc.Row(id="model_container"),
                html.Br(),
            ],
            className='page'
        )


        # Helper functions for callbacks.
        def fetch_params(predictor:object, size:str):
            hyperparameters = predictor.get_hyperparameters()
            if (hyperparameters is not None):
                headers = [html.Th("parameter"), html.Th("value")]
                table_header = [html.Thead(html.Tr(headers), className='thead')]
                rows = [html.Tr([html.Td(k), html.Td(v)]) for k,v in hyperparameters.items()]
                table_body = [html.Tbody(rows)]
                hp_table = dbc.Table(
                    table_header + table_body,
                    dark=True, hover=True, responsive=True,
                    striped=True, bordered=False, className=f"tbl {size} ctr"
                )
            else:
                msg = "Sorry - This model has no parameters."
                hp_table = dbc.Alert(msg, className='alert')
            return hp_table


        """
        - When the page is started and refreshed, this updates things.
        - You can only use objects at the start of the DAG, 
        e.g. not `exp_plot.clickData` because it doesn't exist yet.
        """
        @app.callback(
            [
                Output(component_id="exp_dropdown", component_property="options"),
                Output(component_id="exp_dropdown", component_property="placeholder"),
                Output(component_id="exp_dropdown", component_property="value"),        
                Output(component_id='multitron', component_property='options'),
                Output(component_id='multitron', component_property='placeholder'),
                Output(component_id="multitron", component_property="value"),
            ],
            Input(component_id="initial_load", component_property="n_intervals"),
            [
                State(component_id="exp_dropdown", component_property="value"),        
                State(component_id='multitron', component_property='value'),
            ]
        )
        def refresh(n_intervals:int, queue_id:int, model_ids:int):
            queues = list(orm.Queue)
            if (not queues):
                return [], "None yet", None, [], "None yet", []
            
            # Initially the IDs are None becuse there is no State.
            queues.reverse()
            queue_options = [dict(label=str(q.id), value=q.id) for q in queues]
            if (queue_id is None):
                queue_id = queues[0].id
            
            models = list(orm.Predictor)
            models.reverse()
            model_options = [dict(label=f"Model:{m.id}", value=m.id) for m in models]
            
            return queue_options, None, queue_id, model_options, "Compare models head-to-head", model_ids


        @app.callback(
            Output(component_id="progress_container", component_property="children"),
            Input(component_id="exp_dropdown", component_property="value"),
        )
        def update_progress(queue_id:int):
            if (queue_id is None):
                return None
            queue = orm.Queue.get_by_id(queue_id)
            progress = round(queue.runs_completed/queue.run_count*100)
            
            if (progress<100):
                children = dbc.Progress(
                    id="progress_bar", className='prog_bar ctr', color="secondary",
                    value=progress, label=f"{progress}%"
                )
                return children 
            else:
                return None


        @app.callback(
            [
                Output(component_id='exp_type', component_property='options'),
                Output(component_id='exp_type', component_property='value'),
            ],
            Input(component_id='exp_dropdown', component_property='value'),
            [
                State(component_id='exp_type', component_property='value'),
            ]
        )
        def type_dropdown(queue_id:object, exp_type:str):
            if (queue_id is None):
                raise PreventUpdate
            
            queue = orm.Queue.get_by_id(queue_id)
            analysis_type = queue.algorithm.analysis_type
            if ('classification' in analysis_type):
                score_types = metrics_classify
                if (exp_type not in score_types):
                    exp_type = 'accuracy'
            elif('regression' in analysis_type):   
                score_types = metrics_regress
                if (exp_type not in score_types):
                    exp_type = 'r2'
            
            options = [{"label":m, "value":col} for col, m in score_types.items()]
            return options, exp_type


        @app.callback(
            Output(component_id='exp_plot_contain', component_property='children'),
            Input(component_id='exp_button', component_property='n_clicks'),
            [
                State(component_id='exp_dropdown', component_property='value'),
                State(component_id='exp_type', component_property='value'),
                State(component_id='exp_score', component_property='value'),
                State(component_id='exp_loss', component_property='value'),
            ]
        )
        def plot_experiment(
            n_clicks:int, queue_id:object,
            score_type:str, min_score:float, max_loss:float,
        ):   
            if (queue_id is None):
                queues = list(orm.Queue)
                if (not queues):
                    msg = "Sorry - Cannot display plot because no Queues exist yet. Data will refresh automatically."
                    return dbc.Alert(msg, className='alert')
                else:
                    # Plot the most recent by default
                    queue = queues[-1]
            else:
                queue = orm.Queue.get_by_id(queue_id)
            
            try:
                fig = queue.plot_performance(
                    score_type=score_type, min_score=min_score, max_loss=max_loss,
                    call_display=False, height=500
                )
            except Exception as err_msg:
                return dbc.Alert(str(err_msg), className='alert')
            else:
                fig = dcc.Graph(id="exp_plot", figure=fig, className='exp_plot_contain ctr')
                return fig


        @app.callback(
            Output(component_id='exp_button', component_property='n_clicks'),
            Input(component_id='exp_type', component_property='value'),
            State(component_id='exp_button', component_property='n_clicks'),
        )
        def trigger_graph(score_type:str, n_clicks:int):
            n_clicks += 1
            return n_clicks


        @app.callback(
            Output(component_id='param_pane', component_property='children'),
            # `Input...clickAnnotationData` returns None
            Input(component_id='exp_plot', component_property='clickData'),
        )
        def interactive_params(new_click:dict):
            """
            It's hard to maintain clickData State because exp_plot doesn't exist when 
            the page loads and exp_plot gets overrode when the page refreshes.
            """
            if (new_click is None):
                raise PreventUpdate
            # The docs say to use `json.dumps`, but clickData is just a dict.
            predictor = new_click['points'][0]['customdata'][0]    
            predictor = orm.Predictor.get_by_id(predictor)
            title = f"Model ID: {predictor}"
            model_title = html.P(title, className='header')
            hp_table = fetch_params(predictor, "tbig")
            return [model_title, hp_table]


        @app.callback(
            Output(component_id='model_container', component_property='children'),
            Input(component_id='multitron', component_property='value')
        )
        def model_plots(predictor_ids:list):
            # Initially it's None, but empty list when it's cleared.
            if ((predictor_ids is None) or (not predictor_ids)):
                msg = "Select up to two models from the dropdown above."
                return dbc.Alert(msg, className='alert', style={"width":"50%"})
            pred_count = len(predictor_ids)
            if (pred_count==1):
                col_width = 12
            elif (pred_count==2):
                col_width = 6
            elif (pred_count>2):
                msg = "Sorry - Only 2 models can be displayed at once."
                return dbc.Alert(msg, className='alert')
            multi_cols = []
            for predictor_id in predictor_ids:
                # Only `big_column` is assigned a bootstrap width.
                big_column = []

                predictor = orm.Predictor.get_by_id(predictor_id)
                predictions = list(predictor.predictions)
                if (not predictions):
                    msg = f"Sorry - Metrics for this model are not ready yet. Data will refresh automatically."
                    return dbc.Col(dbc.Alert(msg, className='alert'))
                prediction = predictions[0]

                # === METRICS ===
                metrics = prediction.metrics
                # Need the 'split' to be in the same dict as the metrics.
                metrics_records = []
                for split, metrix in metrics.items():
                    split_dikt = {'split':split}
                    # We want the split to appear first in the dict
                    split_dikt.update(metrix)
                    metrics_records.append(split_dikt)
                cols = list(metrics_records[0].keys())
                headers = [html.Th(c) for c in cols]
                table_header = [html.Thead(html.Tr(headers))]

                metrics_raw = []
                for record in metrics_records:
                    metrics = [v for k,v in record.items()]
                    metrics_raw.append(metrics)

                rows = []
                for cells in metrics_raw:
                    row = html.Tr([html.Td(cell) for cell in cells]) 
                    rows.append(row)            
                table_body = [html.Tbody(rows)]
                metrics_table = dbc.Table(
                    table_header + table_body,
                    dark=True, hover=True, responsive=True, 
                    striped=True, bordered=False, className='tbl tbig ctr'
                )

                row = dbc.Row(
                    dbc.Col(metrics_table, align="center")
                )
                big_column.append(row)
                big_column.append(html.Hr(className='hrz ctr'))

                # === HYPERPARAMETERS ===
                hp_table = fetch_params(predictor, "tsmall")
                row = dbc.Row(
                    dbc.Col(hp_table, align="center")
                )
                big_column.append(row)
                big_column.append(html.Hr(className='hrz ctr'))

                # === LEARNING ===
                learning_curves = predictor.plot_learning_curve(
                    call_display=False, skip_head=False
                )
                learning_curves = [dcc.Graph(figure=fig, className='plots ctr') for fig in learning_curves]
                row = dbc.Row(
                    [
                        dbc.Col(learning_curves, id="learning_plots"),
                    ]
                )
                big_column.append(row)
                big_column.append(html.Hr(className='hrz ctr'))

                # === IMPORTANCE ===
                feature_importance = prediction.feature_importance
                if (feature_importance is not None):
                    prediction = orm.Predictor.get_by_id(predictor_id).predictions[0]
                    content = prediction.plot_feature_importance(top_n=15, call_display=False)
                    content = [dcc.Graph(figure=fig, className='plots ctr') for fig in content]
                elif (feature_importance is None):
                    msg = "Feature importance not calculated for model yet."
                    content = dbc.Alert(msg, className='alert')

                row = dbc.Row(
                    dbc.Col(content, id="importance_plots")
                )
                big_column.append(row)
                big_column.append(html.Hr(className='hrz ctr'))

                # === CLASSIFICATION ===
                analysis_type = predictor.job.queue.algorithm.analysis_type
                if ('classification' in analysis_type):
                    # === ROC ===
                    roc = prediction.plot_roc_curve(call_display=False)
                    roc = dcc.Graph(figure=roc, className='plots ctr')
                    row = dbc.Row(
                        dbc.Col(roc, align="center")
                    )
                    big_column.append(row)
                    big_column.append(html.Hr(className='hrz ctr'))
                    # === PRC ===
                    pr = prediction.plot_precision_recall(call_display=False)
                    pr = dcc.Graph(figure=pr, className='plots ctr')
                    row = dbc.Row(
                        dbc.Col(pr, align="center")
                    )
                    big_column.append(row)
                    big_column.append(html.Hr(className='hrz ctr'))
                    # === CONFUSION ===
                    cms = prediction.plot_confusion_matrix(call_display=False)
                    cms = [dcc.Graph(figure=fig, className='plots ctr') for fig in cms]                
                    row = dbc.Row(
                        dbc.Col(cms, align="center")
                    )
                    big_column.append(row)
                    big_column.append(html.Br())
                    big_column.append(html.Br())
                big_column = dbc.Col(big_column, width=col_width)
                multi_cols.append(big_column)
            return multi_cols

        if (self.server_runtime['mode']=='external'):
            try:
                # Intercept & change Dash's launch message. 
                # Doesn't work in shell, but does work in Jupyter.
                f = io.StringIO()
                with redirect_stdout(f):
                    app.run_server(**self.server_runtime)
                output = f.getvalue()
                # Could check something more advanced, but this works for now.
                if ("Dash app running on" in output):
                    url = f"http://{self.server_runtime['host']}:{self.server_runtime['port']}"
                    print(f"📊 AIQC Tracker {url} 📊")
                else:
                    print(output)
            except:
                self._clean_run = False
                raise
            self._clean_run = True
        else:
            app.run_server(**self.server_runtime)
            self._clean_run = True



"""PyTorch Multi-label Classification with Tabular data"""
# Internal modules
from .. import datum
from ..utils.pytorch import fit
from ..orm import *
# External modules
import torch
import torch.nn as nn
import torchmetrics
from sklearn.preprocessing import OrdinalEncoder, StandardScaler


def fn_build(features_shape, num_classes, **hp):
	model = torch.nn.Sequential(
		nn.Linear(features_shape[0], 12),
		nn.BatchNorm1d(12,12),
		nn.ReLU(),
		nn.Dropout(p=0.5),

		nn.Linear(12, num_classes),
		nn.Softmax(dim=1),
	)
	return model


def fn_lose(**hp):
	loser = nn.CrossEntropyLoss(reduction=hp['reduction'])
	return loser


def fn_train(model, loser, optimizer, samples_train, samples_evaluate, **hp):
	return fit(
		model, loser, optimizer, 
		samples_train, samples_evaluate,
		epochs=10, batch_size=hp['batch_size'],  
		metrics=[torchmetrics.Accuracy(),torchmetrics.F1Score()]
	)

def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=3):
	if (fold_count is not None):
		file_path = datum.get_path('iris_10x.tsv')
	else:
		file_path = datum.get_path('iris.tsv')

	d_id = Dataset.Tabular.from_path(
		file_path = file_path
		, source_file_format = 'tsv'
		, dtype = None
	).id
	
	label_column = 'species'
	l_id = Label.from_dataset(dataset_id=d_id, columns=[label_column]).id
	f_id = Feature.from_dataset(dataset_id=d_id, exclude_columns=[label_column]).id

	if (fold_count is not None):
		size_test = 0.25
		size_validation = None
	elif (fold_count is None):
		size_test = 0.18
		size_validation = 0.14

	s_id = Splitset.make(
		feature_ids = [f_id]
		, label_id = l_id
		, size_test = size_test
		, size_validation = size_validation
	).id

	if (fold_count is not None):
		fs_id = Foldset.from_splitset(splitset_id=s_id, fold_count=fold_count).id
	else:
		fs_id = None

	LabelCoder.from_label(label_id=l_id, sklearn_preprocess=OrdinalEncoder())

	e_id = Encoderset.from_feature(feature_id=f_id).id
	FeatureCoder.from_encoderset(
		encoderset_id = e_id
		, sklearn_preprocess = StandardScaler(copy=False)
		, dtypes = ['float64']
	)

	a_id = Algorithm.make(
		library = "pytorch"
		, analysis_type = "classification_multi"
		, fn_build = fn_build
		, fn_train = fn_train
        , fn_lose = fn_lose
	).id

	hyperparameters = {
		"reduction":['mean'], "batch_size":[5]
	}
	h_id = Hyperparamset.from_algorithm(
		algorithm_id=a_id, hyperparameters=hyperparameters
	).id

	queue = Queue.from_algorithm(
		algorithm_id = a_id
		, splitset_id = s_id
		, foldset_id = fs_id
		, hyperparamset_id = h_id
		, repeat_count = repeat_count
		, permute_count = permute_count
	)
	return queue




from .wrangle import if_1d_make_2d, cols_by_indices, colIndices_from_colNames
import inspect, warnings
from scipy import sparse
from textwrap import dedent
import numpy as np


# Used during encoding validation.
categorical_encoders = [
	'OneHotEncoder', 'LabelEncoder', 'OrdinalEncoder', 
	'Binarizer', 'LabelBinarizer', 'MultiLabelBinarizer'
]


# Used by `sklearn.preprocessing.FunctionTransformer` to normalize images.
def div255(x): return x/255
def mult255(x): return x*255


def check_sklearn_attributes(sklearn_preprocess:object, is_label:bool):
	#This function is used by Featurecoder too, so don't put label-specific things in here.
	if (inspect.isclass(sklearn_preprocess)):
		raise Exception(dedent("""
			Yikes - The encoder you provided is a class name, but it should be a class instance.\n
			Class (incorrect): `OrdinalEncoder`
			Instance (correct): `OrdinalEncoder()`
		"""))

	# Encoder parent modules vary: `sklearn.preprocessing._data` vs `sklearn.preprocessing._label`
	# Feels cleaner than this: https://stackoverflow.com/questions/14570802/python-check-if-object-is-instance-of-any-class-from-a-certain-module
	coder_type = str(type(sklearn_preprocess))
	if ('sklearn.preprocessing' not in coder_type):
		raise Exception(dedent("""
			Yikes - At this point in time, only `sklearn.preprocessing` encoders are supported.
			https://scikit-learn.org/stable/modules/classes.html#module-sklearn.preprocessing
		"""))
	elif ('sklearn.preprocessing' in coder_type):
		if (not hasattr(sklearn_preprocess, 'fit')):    
			raise Exception(dedent("""
				Yikes - The `sklearn.preprocessing` method you provided does not have a `fit` method.\n
				Please use one of the uppercase methods instead.
				For example: use `RobustScaler` instead of `robust_scale`.
			"""))

		if (hasattr(sklearn_preprocess, 'sparse')):
			if (sklearn_preprocess.sparse == True):
				try:
					sklearn_preprocess.sparse = False
					print(dedent("""
						=> Info - System overriding user input to set `sklearn_preprocess.sparse=False`.
							This would have generated 'scipy.sparse.csr.csr_matrix', causing Keras training to fail.
					"""))
				except:
					raise Exception(dedent(f"""
						Yikes - Detected `sparse==True` attribute of {sklearn_preprocess}.
						System attempted to override this to False, but failed.
						FYI `sparse` is True by default if left blank.
						This would have generated 'scipy.sparse.csr.csr_matrix', causing Keras training to fail.\n
						Please try again with False. For example, `OneHotEncoder(sparse=False)`.
					"""))

		if (hasattr(sklearn_preprocess, 'drop')):
			if (sklearn_preprocess.drop is not None):
				try:
					sklearn_preprocess.drop = None
					print(dedent("""
						=> Info - System overriding user input to set `sklearn_preprocess.drop`.
							System cannot handle `drop` yet when dynamically inverse_transforming predictions.
					"""))
				except:
					raise Exception(dedent(f"""
						Yikes - Detected `drop is not None` attribute of {sklearn_preprocess}.
						System attempted to override this to None, but failed.
					"""))

		if (hasattr(sklearn_preprocess, 'copy')):
			if (sklearn_preprocess.copy == True):
				try:
					sklearn_preprocess.copy = False
					print(dedent("""
						=> Info - System overriding user input to set `sklearn_preprocess.copy=False`.
							This saves memory when concatenating the output of many encoders.
					"""))
				except:
					raise Exception(dedent(f"""
						Yikes - Detected `copy==True` attribute of {sklearn_preprocess}.
						System attempted to override this to False, but failed.
						FYI `copy` is True by default if left blank, which consumes memory.\n
						Please try again with 'copy=False'.
						For example, `StandardScaler(copy=False)`.
					"""))
		
		if (hasattr(sklearn_preprocess, 'sparse_output')):
			if (sklearn_preprocess.sparse_output == True):
				try:
					sklearn_preprocess.sparse_output = False
					print(dedent("""
						=> Info - System overriding user input to set `sklearn_preprocess.sparse_output=False`.
							This would have generated 'scipy.sparse.csr.csr_matrix', causing Keras training to fail.
					"""))
				except:
					raise Exception(dedent(f"""
						Yikes - Detected `sparse_output==True` attribute of {sklearn_preprocess}.
						System attempted to override this to True, but failed.
						Please try again with 'sparse_output=False'.
						This would have generated 'scipy.sparse.csr.csr_matrix', causing Keras training to fail.\n
						For example, `LabelBinarizer(sparse_output=False)`.
					"""))

		if (hasattr(sklearn_preprocess, 'order')):
			if (sklearn_preprocess.order == 'F'):
				try:
					sklearn_preprocess.order = 'C'
					print(dedent("""
						=> Info - System overriding user input to set `sklearn_preprocess.order='C'`.
							This changes the output shape of the 
					"""))
				except:
					raise Exception(dedent(f"""
						System attempted to override this to 'C', but failed.
						Yikes - Detected `order=='F'` attribute of {sklearn_preprocess}.
						Please try again with 'order='C'.
						For example, `PolynomialFeatures(order='C')`.
					"""))

		if (hasattr(sklearn_preprocess, 'encode')):
			if (sklearn_preprocess.encode == 'onehot'):
				# Multiple options here, so don't override user input.
				raise Exception(dedent(f"""
					Yikes - Detected `encode=='onehot'` attribute of {sklearn_preprocess}.
					FYI `encode` is 'onehot' by default if left blank and it predictors in 'scipy.sparse.csr.csr_matrix',
					which causes Keras training to fail.\n
					Please try again with 'onehot-dense' or 'ordinal'.
					For example, `KBinsDiscretizer(encode='onehot-dense')`.
				"""))

		if (
			(is_label==True)
			and
			(not hasattr(sklearn_preprocess, 'inverse_transform'))
		):
			print(dedent("""
				Warning - The following encoders do not have an `inverse_transform` method.
				It is inadvisable to use them to encode Labels during training, 
				because you may not be able to programmatically decode your raw predictions 
				when it comes time for inference (aka non-training predictions):

				[Binarizer, KernelCenterer, Normalizer, PolynomialFeatures, SplineTransformer]
			"""))

		"""
		- Binners like 'KBinsDiscretizer' and 'QuantileTransformer'
			will place unseen observations outside bounds into existing min/max bin.
		- I assume that someone won't use a custom FunctionTransformer, for categories
			when all of these categories are available.
		- LabelBinarizer is not threshold-based, it's more like an OHE.
		"""
		only_fit_train = True
		stringified_coder = str(sklearn_preprocess)
		is_categorical = False
		for c in categorical_encoders:
			if (stringified_coder.startswith(c)):
				only_fit_train = False
				is_categorical = True
				break

		return sklearn_preprocess, only_fit_train, is_categorical


def fit_dynamicDimensions(sklearn_preprocess:object, samples_to_fit:object):
	"""
	- There are 17 uppercase sklearn encoders, and 10 different data types across float, str, int 
		when consider negatives, 2D multiple columns, 2D single columns.
	- Different encoders work with different data types and dimensionality.
	- This function normalizes that process by coercing the dimensionality that the encoder wants,
		and erroring if the wrong data type is used. The goal in doing so is to return 
		that dimensionality for future use.

	- `samples_to_transform` is pre-filtered for the appropriate `matching_columns`.
	- The rub lies in that if you have many columns, but the encoder only fits 1 column at a time, 
		then you return many fits for a single type of preprocess.
	- Remember this is for a single Featurecoder that is potential returning multiple fits.

	- UPDATE: after disabling LabelBinarizer and LabelEncoder from running on multiple columns,
		everything seems to be fitting as "2D_multiColumn", but let's keep the logic for new sklearn methods.
	"""
	fitted_encoders = []
	incompatibilities = {
		"string": [
			"KBinsDiscretizer", "KernelCenterer", "MaxAbsScaler", 
			"MinMaxScaler", "PowerTransformer", "QuantileTransformer", 
			"RobustScaler", "StandardScaler"
		]
		, "float": ["LabelBinarizer"]
		, "numeric array without dimensions both odd and square (e.g. 3x3, 5x5)": ["KernelCenterer"]
	}

	with warnings.catch_warnings(record=True) as w:
		try:
			#`samples_to_fit` is coming in as 2D.
			# Remember, we are assembling `fitted_encoders` dict, not accesing it.
			fit_encoder = sklearn_preprocess.fit(samples_to_fit)
			fitted_encoders.append(fit_encoder)
		except:
			# At this point, "2D" failed. It had 1 or more columns.
			try:
				width = samples_to_fit.shape[1]
				if (width > 1):
					# Reshape "2D many columns" to “3D of 2D single columns.”
					samples_to_fit = samples_to_fit[None].T                    
					# "2D single column" already failed. Need it to fail again to trigger except.
				elif (width == 1):
					# Reshape "2D single columns" to “3D of 2D single columns.”
					samples_to_fit = samples_to_fit.reshape(1, samples_to_fit.shape[0], 1)    
				# Fit against each 2D array within the 3D array.
				for i, arr in enumerate(samples_to_fit):
					fit_encoder = sklearn_preprocess.fit(arr)
					fitted_encoders.append(fit_encoder)
			except:
				# At this point, "2D single column" has failed.
				try:
					# So reshape the "3D of 2D_singleColumn" into "2D of 1D for each column."
					# This transformation is tested for both (width==1) as well as (width>1). 
					samples_to_fit = samples_to_fit.transpose(2,0,1)[0]
					# Fit against each column in 2D array.
					for i, arr in enumerate(samples_to_fit):
						fit_encoder = sklearn_preprocess.fit(arr)
						fitted_encoders.append(fit_encoder)
				except:
					raise Exception(dedent(f"""
						Yikes - Encoder failed to fit the columns you filtered.\n
						Either the data is dirty (e.g. contains NaNs),
						or the encoder might not accept negative values (e.g. PowerTransformer.method='box-cox'),
						or you used one of the incompatible combinations of data type and encoder seen below:\n
						{incompatibilities}
					"""))
				else:
					encoding_dimension = "1D"
			else:
				encoding_dimension = "2D_singleColumn"
		else:
			encoding_dimension = "2D_multiColumn"
	return fitted_encoders, encoding_dimension


def transform_dynamicDimensions(
	fitted_encoders:list
	, encoding_dimension:str
	, samples_to_transform:object
):
	"""
	- UPDATE: after disabling LabelBinarizer and LabelEncoder from running on multiple columns,
		everything seems to be fitting as "2D_multiColumn", but let's keep the logic for new sklearn methods.
	"""
	if (encoding_dimension == '2D_multiColumn'):
		# Our `to_numpy` method fetches data as 2D. So it has 1+ columns. 
		encoded_samples = fitted_encoders[0].transform(samples_to_transform)
		encoded_samples = if_1d_make_2d(array=encoded_samples)
	elif (encoding_dimension == '2D_singleColumn'):
		# Means that `2D_multiColumn` arrays cannot be used as is.
		width = samples_to_transform.shape[1]
		if (width == 1):
			# It's already "2D_singleColumn"
			encoded_samples = fitted_encoders[0].transform(samples_to_transform)
			encoded_samples = if_1d_make_2d(array=encoded_samples)
		elif (width > 1):
			# Data must be fed into encoder as separate '2D_singleColumn' arrays.
			# Reshape "2D many columns" to “3D of 2D singleColumns” so we can loop on it.
			encoded_samples = samples_to_transform[None].T
			encoded_arrs = []
			for i, arr in enumerate(encoded_samples):
				encoded_arr = fitted_encoders[i].transform(arr)
				encoded_arr = if_1d_make_2d(array=encoded_arr)  
				encoded_arrs.append(encoded_arr)
			encoded_samples = np.array(encoded_arrs).T

			# From "3D of 2Ds" to "2D wide"
			# When `encoded_samples` was accidentally a 3D shape, this fixed it:
			"""
			if (len(encoded_samples.shape) == 3):
				encoded_samples = encoded_samples.transpose(
					1,0,2
				).reshape(
					# where index represents dimension.
					encoded_samples.shape[1],
					encoded_samples.shape[0]*encoded_samples.shape[2]
				)
			"""
			del encoded_arrs
	elif (encoding_dimension == '1D'):
		# From "2D_multiColumn" to "2D with 1D for each column"
		# This `.T` works for both single and multi column.
		encoded_samples = samples_to_transform.T
		# Since each column is 1D, we care about rows now.
		length = encoded_samples.shape[0]
		if (length == 1):
			#encoded_samples = fitted_encoders[0].transform(encoded_samples)
			# to get text feature_extraction working.
			encoded_samples = fitted_encoders[0].transform(encoded_samples[0])
			# Some of these 1D encoders also output 1D.
			# Need to put it back into 2D.
			encoded_samples = if_1d_make_2d(array=encoded_samples)  
		elif (length > 1):
			encoded_arrs = []
			for i, arr in enumerate(encoded_samples):
				encoded_arr = fitted_encoders[i].transform(arr)
				# Check if it is 1D before appending.
				encoded_arr = if_1d_make_2d(array=encoded_arr)              
				encoded_arrs.append(encoded_arr)
			# From "3D of 2D_singleColumn" to "2D_multiColumn"
			encoded_samples = np.array(encoded_arrs).T
			del encoded_arrs
	if (sparse.issparse(encoded_samples)):
		return encoded_samples.todense()
	return encoded_samples


def encoder_fit_labels(
	arr_labels:object, samples_train:list,
	labelcoder:object
):
	"""
	- All Label columns are always used during encoding.
	- Rows determine what fit happens.
	"""
	if (labelcoder is not None):
		preproc = labelcoder.sklearn_preprocess

		if (labelcoder.only_fit_train == True):
			labels_to_fit = arr_labels[samples_train]
		elif (labelcoder.only_fit_train == False):
			labels_to_fit = arr_labels
			
		fitted_coders, encoding_dimension = fit_dynamicDimensions(
			sklearn_preprocess = preproc
			, samples_to_fit = labels_to_fit
		)
		# Save the fit.
		fitted_encoders = fitted_coders[0]#take out of list before adding to dict.
	return fitted_encoders


def encoder_transform_labels(
	arr_labels:object,
	fitted_encoders:object, labelcoder:object 
):
	encoding_dimension = labelcoder.encoding_dimension
	
	arr_labels = transform_dynamicDimensions(
		fitted_encoders = [fitted_encoders] # `list(fitted_encoders)`, fails.
		, encoding_dimension = encoding_dimension
		, samples_to_transform = arr_labels
	)
	return arr_labels


def encoderset_fit_features(
	arr_features:object, samples_train:list,
	encoderset:object,
):
	featurecoders = list(encoderset.featurecoders)
	fitted_encoders = []
	if (len(featurecoders) > 0):
		f_cols = encoderset.feature.columns
		
		# For each featurecoder: fetch, transform, & concatenate matching features.
		# One nested list per Featurecoder. List of lists.
		for featurecoder in featurecoders:
			preproc = featurecoder.sklearn_preprocess

			if (featurecoder.only_fit_train == True):
				features_to_fit = arr_features[samples_train]
			elif (featurecoder.only_fit_train == False):
				features_to_fit = arr_features
			
			# Handles `Dataset.Sequence` by stacking the 2D arrays into a tall 2D array.
			features_shape = features_to_fit.shape
			if (len(features_shape)==3):
				rows_2D = features_shape[0] * features_shape[1]
				features_to_fit = features_to_fit.reshape(rows_2D, features_shape[2])
			elif (len(features_shape)==4):
				rows_2D = features_shape[0] * features_shape[1] * features_shape[2]
				features_to_fit = features_to_fit.reshape(rows_2D, features_shape[3])

			# Only fit these columns.
			matching_columns = featurecoder.matching_columns
			# Get the indices of the desired columns.
			col_indices = colIndices_from_colNames(
				column_names=f_cols, desired_cols=matching_columns
			)
			# Filter the array using those indices.
			features_to_fit = cols_by_indices(features_to_fit, col_indices)

			# Fit the encoder on the subset.
			fitted_coders, encoding_dimension = fit_dynamicDimensions(
				sklearn_preprocess = preproc
				, samples_to_fit = features_to_fit
			)
			fitted_encoders.append(fitted_coders)
	return fitted_encoders


def encoderset_transform_features(
	arr_features:object,
	fitted_encoders:list, encoderset:object 
):
	"""
	- Can't overwrite columns with data of different type (e.g. encoding object to int), 
		so they have to be pieced together.
	"""
	featurecoders = list(encoderset.featurecoders)
	if (len(featurecoders) > 0):
		# Handle Sequence (part 1): reshape 3D to tall 2D for transformation.
		og_shape = arr_features.shape
		if (len(og_shape)==3):
			rows_2D = og_shape[0] * og_shape[1]
			arr_features = arr_features.reshape(rows_2D, og_shape[2])
		elif (len(og_shape)==4):
			rows_2D = og_shape[0] * og_shape[1] * og_shape[2]
			arr_features = arr_features.reshape(rows_2D, og_shape[3])

		f_cols = encoderset.feature.columns
		transformed_features = None #Used as a placeholder for `np.concatenate`.
		for featurecoder in featurecoders:
			idx = featurecoder.index
			fitted_coders = fitted_encoders[idx]# returns list
			encoding_dimension = featurecoder.encoding_dimension
			
			# Only transform these columns.
			matching_columns = featurecoder.matching_columns
			# Get the indices of the desired columns.
			col_indices = colIndices_from_colNames(
				column_names=f_cols, desired_cols=matching_columns
			)
			# Filter the array using those indices.
			features_to_transform = cols_by_indices(arr_features, col_indices)

			if (idx == 0):
				# It's the first encoder. Nothing to concat with, so just overwite the None value.
				transformed_features = transform_dynamicDimensions(
					fitted_encoders = fitted_coders
					, encoding_dimension = encoding_dimension
					, samples_to_transform = features_to_transform
				)
			elif (idx > 0):
				encoded_features = transform_dynamicDimensions(
					fitted_encoders = fitted_coders
					, encoding_dimension = encoding_dimension
					, samples_to_transform = features_to_transform
				)
				# Then concatenate w previously encoded features.
				transformed_features = np.concatenate(
					(transformed_features, encoded_features)
					, axis = 1
				)
		
		# After all featurecoders run, merge in leftover, unencoded columns.
		leftover_columns = featurecoders[-1].leftover_columns
		if (len(leftover_columns) > 0):
			# Get the indices of the desired columns.
			col_indices = colIndices_from_colNames(
				column_names=f_cols, desired_cols=leftover_columns
			)
			# Filter the array using those indices.
			leftover_features = cols_by_indices(arr_features, col_indices)
					
			transformed_features = np.concatenate(
				(transformed_features, leftover_features)
				, axis = 1
			)
		# Handle Sequence (part 2): reshape tall 2D back to 3D.
		# This checks `==3` intentionaly!!!
		if (len(og_shape)==3):
			transformed_features = arr_features.reshape(
				og_shape[0],
				og_shape[1],
				og_shape[2]
			)
		elif(len(og_shape)==4):
			transformed_features = arr_features.reshape(
				og_shape[0],
				og_shape[1],
				og_shape[2],
				og_shape[3]
			)
			
	elif (len(featurecoders) == 0):
		transformed_features = arr_features
	return transformed_features



"""
Plots
└── Documentation = https://aiqc.readthedocs.io/en/latest/notebooks/visualization.html
└── Data is prepared in the `plot_*` methods of the Queue and Predictor ORM classes.
"""
import plotly.graph_objects as go
import plotly.express as px
import plotly.figure_factory as ff
from re import sub


def set_dash_font(fig:object):
	"""Dash knows about `/lab/assets` for Abel .tff, but regular charts don't"""
	return fig.update_layout(
		font=dict(family="Abel, Avenir", size=14)
		, titlefont=dict(family="Abel, Avenir")
		, hoverlabel=dict(font=dict(family="Abel, Avenir"))
	)


class Plot(object):
	def __init__(self):
		"""
		- Defines a template that is passed to the other plots.
		- Font gets overriden to Abel below by Dash.
		"""
		self.plot_template = dict(
            layout=go.Layout(
			    font=dict(family="Avenir", color='#FAFAFA', size=13)
                , title=dict(x=0.05, y=0.95)
                , titlefont=dict(family="Avenir")
                , title_pad=dict(b=50, t=20)
                , plot_bgcolor='#182d41'
                , paper_bgcolor='#182d41'
                , hovermode='closest'
                , hoverlabel=dict(
                    bgcolor="#122536"
                    , font=dict(family="Avenir", size=15)
                )
            )
        )


	def performance(
		self, dataframe:object, score_type:str, 
		score_display:str, call_display:bool=True,
		height:int=None
	):
		fig = px.line(
			dataframe
			, title = 'Models Metrics by Split'
			, x = 'loss'
			, y = score_type
			, color = 'predictor_id'
			, height = height
			, hover_data = ['predictor_id', 'split', 'loss', score_type]
			, line_shape='spline'
		)
		fig.update_traces(
			mode = 'markers+lines'
			, line = dict(width = 2)
			, marker = dict(
				size=8, line=dict(width=2, color='white')
			)
		)
		fig.update_layout(
			title_x=0.5#center
			, xaxis_title = "Loss"
			, yaxis_title = score_display
			, template = self.plot_template
			, showlegend=False#gets way too busy
		)

		fig.update_xaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
		fig.update_yaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
		
		if (call_display==True):
			fig.show()
		else:
			fig = set_dash_font(fig)
			return fig


	def learning_curve(
		self, dataframe:object, history_pairs:dict,
		skip_head:bool=False, call_display:bool=True
	):
		"""Dataframe rows are epochs and columns are metric names."""
		if (skip_head==True):
			dataframe = dataframe.tail(round(dataframe.shape[0]*.85))
		# Spline seems to crash with too many points.
		if (dataframe.shape[0] >= 400):
			line_shape = 'linear'
		elif (dataframe.shape[0] < 400):
			line_shape = 'spline'

		# Create a plot for every pair.
		figs = []
		for train,val in history_pairs.items():
			metric_name = sub("_"," ",train[6:])
			df = dataframe[[train,val]]
			# The y axis is so variable that we need a range for it
			vals_all = list(df[train]) + list(df[val])
			tick_size = abs((max(vals_all) - min(vals_all))/7)
			fig = px.line(
				df, title=f"Training History: {metric_name}", line_shape=line_shape
			)
			fig.update_layout(
				title_y = 1
				, xaxis_title = "epochs"
				, yaxis_title = metric_name
				, legend_title = None
				, height = 400
				, template = self.plot_template
				, yaxis = dict(
					side = "right"
					, tickmode = 'linear'
					#when high loss, can't small fixed dtick. it freezes browser.
					# `nticks` neither working here nor in `update_yaxes` below
					, dtick = tick_size
				)
				, legend = dict(
					orientation="h"
					, yanchor="bottom"
					, y=1.02
					, xanchor="right"
					, x=1
				)
			)
			fig.update_xaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
			fig.update_yaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
			if (call_display==True):
				fig.show()
			else:
				fig = set_dash_font(fig)
				figs.append(fig)
		if (call_display==False):
			return figs


	def confusion_matrix(self, cm_by_split, labels, call_display:bool=True):
		figs = []
		for split, cm in cm_by_split.items():
			# change each element of z to type string for annotations
			cm_text = [[str(y) for y in x] for x in cm]

			fig = ff.create_annotated_heatmap(
				cm, x=labels, y=labels
				, annotation_text=cm_text
				, colorscale=px.colors.sequential.BuGn
				, showscale=False
				, colorbar={"title": 'Count'})

			# add custom xaxis title
			fig.add_annotation(dict(
				font=dict(color="white", size=15)
				, x=0.5
				, y=1.3
				, showarrow=False
				, text="Predicted Label"
				, xref="paper"
				, yref="paper"
			))

			# add custom yaxis title
			fig.add_annotation(dict(
				font=dict(color="white", size=15)
				, x = -0.19, y = 0.5
				# https://plotly.com/python/figure-structure/#positioning-with-paper-container-coordinates-or-axis-domain-coordinates
				, xref="x domain", yref="paper"
				, showarrow=False
				, text="Actual Label"
				, textangle=-90

			))

			fig.update_layout(
				title = dict(
					text = f"Confusion Matrix: {split.capitalize()}"
					, y = 0.0
					, x = 0.55
					, xanchor = 'center'
					, yanchor = 'bottom'
				)
				, template=self.plot_template
				, height=375  # if too small, it won't render in Jupyter.
				, yaxis=dict(
					tickmode='linear'
					, tick0=0.0
					, dtick=1.0
					, fixedrange=True#prevents zoom/pan
					, tickfont = dict(size=13)
				)
				, xaxis=dict(
					categoryorder = 'category descending'
					, fixedrange = True#prevents zoom/pan
					, tickfont = dict(size=13)
				)
				#, margin = dict(r=75, l=150)#impacts y axis annotation x position
				, margin = dict(l=200)#impacts y axis annotation x position
			)

			fig.update_traces(
				hovertemplate = """predicted: %{x}<br>actual: %{y}<br>count: %{z}<extra></extra>"""
			)
			if (call_display==True):
				fig.show()
			else:
				fig = set_dash_font(fig)
				figs.append(fig)
		if (call_display==False):
			return figs


	def precision_recall(self, dataframe:object, call_display:bool=True):
		fig = px.line(
			dataframe
			, x = 'recall'
			, y = 'precision'
			, color = 'split'
			, title = 'Precision-Recall Curves'
		)
		fig.update_layout(
			legend_title = None
			, template = self.plot_template
			, height = 500
			, xaxis = dict(title='Recall (Tₚ / Tₚ+Fₚ)')
			, yaxis = dict(
				title = 'Precision (Tₚ / Tₚ+Fₙ)'
				, side = "right"
				, tickmode = 'linear'
				, tick0 = 0.0
				, dtick = 0.05
			)
			, legend = dict(
				orientation="h"
				, yanchor="bottom"
				, y=1.02
				, xanchor="right"
				, x=1
			)
		)
		fig.update_xaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
		fig.update_yaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
		if (call_display==True):
			fig.show()
		else:
			fig = set_dash_font(fig)
			return fig


	def roc_curve(self, dataframe:object, call_display:bool=True):
		fig = px.line(
			dataframe
			, x = 'fpr'
			, y = 'tpr'
			, color = 'split'
			, title = 'Receiver Operating Characteristic (ROC) Curves'
		)
		fig.update_layout(
			legend_title = None
			, template = self.plot_template
			, height = 500
			, xaxis = dict(
				title = "False Positive Rate (FPR)"
				, tick0 = 0.00
				, range = [-0.025,1]
			)
			, yaxis = dict(
				title = "True Positive Rate (TPR)"
				, side = "left"
				, tickmode = 'linear'
				, tick0 = 0.00
				, dtick = 0.05
				, range = [0,1.05]
			)
			, legend = dict(
				orientation="h"
				, yanchor="bottom"
				, y=1.02
				, xanchor="right"
				, x=1
			)
			, shapes=[
				dict(
					type = 'line'
					, y0=0, y1=1
					, x0=0, x1=1
					, line = dict(dash='dot', width=2, color='#3b4043')
			)]
		)
		fig.update_xaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
		fig.update_yaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
		if (call_display==True):
			fig.show()
		else:
			fig = set_dash_font(fig)
			return fig
	

	def feature_importance(
		self, feature_impacts:object, feature_id:int,
		permute_count:int, top_n:int,
		height:int, margin_left:int,
		call_display:bool
	):
		if (top_n is not None):
			title = f"Feature Importance <sub>(feature.id:{feature_id}, permute_count:{permute_count}, top_n:{top_n})</sub><br><br>"
		elif (top_n is None):
			title = f"Feature Importance <sub>(feature.id:{feature_id}, permute_count:{permute_count})</sub><br><br>"
		
		fig = go.Figure()
		for feature, impacts in feature_impacts.items():
			fig.add_trace(go.Box(x=impacts, name=feature))
		fig.update_layout(
			template = self.plot_template
			, height = height
			, title = title
			, showlegend = False
			, margin = dict(l=margin_left)
		)
		fig.update_xaxes(
			title = f"Importance<br><sup>[permuted column loss - training loss]</sup>",
			# ticks not really showing.
			tickangle=45, nticks=15, gridcolor='#2c3c4a'
		)
		if (call_display==True):
			fig.show()
		else:
			fig = set_dash_font(fig)
			return fig



"""TensorFlow Regression with Tabular data"""
# Internal modules
from .. import datum
from ..orm import *
# External modules
import tensorflow as tf
import tensorflow.keras.layers as l
from sklearn.preprocessing import PowerTransformer, MinMaxScaler, OrdinalEncoder
import numpy as np


def fn_build(features_shape, label_shape, **hp):
	m = tf.keras.models.Sequential()
	m.add(l.Dense(units=hp['neuron_count'], kernel_initializer='normal', activation='relu'))
	m.add(l.Dropout(0.15))
	m.add(l.Dense(units=hp['neuron_count'], kernel_initializer='normal', activation='relu'))
	m.add(l.Dense(units=label_shape[0], kernel_initializer='normal'))
	return m


def fn_optimize(**hp):
	optimizer = tf.keras.optimizers.RMSprop()
	return optimizer


def fn_train(model, loser, optimizer, samples_train, samples_evaluate, **hp):
	model.compile(
		loss=loser
		, optimizer=optimizer
		, metrics = ['mean_squared_error']
	)

	model.fit(
		samples_train['features'], samples_train['labels']
		, validation_data = (
			samples_evaluate['features'],
			samples_evaluate['labels'])
		, verbose = 0
		, batch_size = 3
		, epochs = hp['epochs']
		, callbacks = [tf.keras.callbacks.History()]
	)
	return model


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=3):
	hyperparameters = {
		"neuron_count": [24]
		, "epochs": [10]
	}

	df = datum.to_pandas('houses.csv')
	# testing Labelpolater (we don't have a regression-sequence example yet).
	df['price'][0] = np.NaN
	df['price'][5] = np.NaN
	df['price'][10] = np.NaN
	# testing Featurepolater 2D.
	df['nox'][5] = np.NaN
	df['indus'][10] = np.NaN
	df['age'][19] = np.NaN
	
	d_id = Dataset.Tabular.from_pandas(dataframe=df).id
	
	label_column = 'price'
	l_id = Label.from_dataset(dataset_id=d_id, columns=[label_column]).id
	
	LabelInterpolater.from_label(
		label_id = l_id
		, interpolate_kwargs = dict(
			method = 'linear'
			, limit_direction = 'both'
			, limit_area = None
			, axis = 0
			, order = 1
		)
	)
	LabelCoder.from_label(
		label_id = l_id
		, sklearn_preprocess = PowerTransformer(method='box-cox', copy=False)
	)

	f_id = Feature.from_dataset(dataset_id=d_id, exclude_columns=[label_column]).id
	
	i_id = Interpolaterset.from_feature(feature_id=f_id).id
	FeatureInterpolater.from_interpolaterset(interpolaterset_id=i_id, columns='nox')
	FeatureInterpolater.from_interpolaterset(interpolaterset_id=i_id, dtypes='float64')
	
	e_id = Encoderset.from_feature(feature_id=f_id).id
	FeatureCoder.from_encoderset(
		encoderset_id=e_id
		, include = False
		, dtypes = ['int64']
		, sklearn_preprocess = MinMaxScaler(copy=False)
	)
	# Expect double None (dtypes,columns) to use all columns because nothing is excluded.
	FeatureCoder.from_encoderset(
		encoderset_id=e_id
		, include = False
		, dtypes = None
		, columns = None
		, sklearn_preprocess = OrdinalEncoder()
	)

	if (fold_count is not None):
		size_test = 0.25
		size_validation = None
	elif (fold_count is None):
		size_test = 0.18
		size_validation = 0.14

	s_id = Splitset.make(
		feature_ids = [f_id]
		, label_id = l_id
		, size_test = size_test
		, size_validation = size_validation
		, bin_count = 5
	).id

	if (fold_count is not None):
		fs_id = Foldset.from_splitset(
			splitset_id = s_id
			, fold_count = fold_count
			, bin_count = 3
		).id
	else:
		fs_id = None

	a_id = Algorithm.make(
		library = "keras"
		, analysis_type = "regression"
		, fn_build = fn_build
		, fn_train = fn_train
		, fn_optimize = fn_optimize
	).id

	h_id = Hyperparamset.from_algorithm(
		algorithm_id=a_id, hyperparameters=hyperparameters
	).id

	queue = Queue.from_algorithm(
		algorithm_id = a_id
		, splitset_id = s_id
		, foldset_id = fs_id
		, hyperparamset_id = h_id
		, repeat_count = repeat_count
		, permute_count = permute_count
	)
	return queue



from re import sub
from sklearn import metrics
import numpy as np


# Used by plots and UI.
metrics_classify = dict(
	accuracy='Accuracy',
	f1='F1',
	roc_auc='ROC-AUC',
	precision='Precision', 
	recall='Recall',
)
metrics_classify_cols = list(metrics_classify.keys())
metrics_regress = dict(
	r2='R²',
	mse='MSE',
	explained_variance='ExpVar',
)
metrics_regress_cols = list(metrics_regress.keys())
metrics_all = {**metrics_classify, **metrics_regress}


def display_name(score_type:str):
	#`score_type` accesses df column, whereas `score_display` displays in plot
	score_display = sub("_", " ", score_type)
	if (score_display == "r2"):
		score_display = "R²"
	elif ((score_display=="roc auc") or (score_display=="mse")):
		score_display = score_display.upper()
	else:
		score_display = score_display.title()
	return score_display


def split_classification_metrics(labels_processed, predictions, probabilities, analysis_type):
	"""
		- Be sure to register any new metrics in `metrics_classify` global.
		- Very rarely, these still fail (e.g. ROC when only 1 class of label is predicted).
	"""
	if (analysis_type == "classification_binary"):
		average = "binary"
		roc_average = "micro"
		roc_multi_class = None
	elif (analysis_type == "classification_multi"):
		average = "weighted"
		roc_average = "weighted"
		roc_multi_class = "ovr"
		
	split_metrics = {}		
	# Let the classification_multi labels hit this metric in OHE format.
	split_metrics['roc_auc'] = metrics.roc_auc_score(labels_processed, probabilities, average=roc_average, multi_class=roc_multi_class)
	# Then convert the classification_multi labels ordinal format.
	if (analysis_type == "classification_multi"):
		labels_processed = np.argmax(labels_processed, axis=1)

	split_metrics['accuracy'] = metrics.accuracy_score(labels_processed, predictions)
	split_metrics['precision'] = metrics.precision_score(labels_processed, predictions, average=average, zero_division=0)
	split_metrics['recall'] = metrics.recall_score(labels_processed, predictions, average=average, zero_division=0)
	split_metrics['f1'] = metrics.f1_score(labels_processed, predictions, average=average, zero_division=0)
	return split_metrics


def split_regression_metrics(data, predictions):
	"""Be sure to register any new metrics in `metrics_regress` global."""
	split_metrics = {}
	data_shape = data.shape
	# Unsupervised sequences and images have many data points for a single sample.
	# These metrics only work with 2D data, and all we are after is comparing each number to the real number.
	if (len(data_shape) == 5):
		data = data.reshape(data_shape[0]*data_shape[1]*data_shape[2]*data_shape[3], data_shape[4])
		predictions = predictions.reshape(data_shape[0]*data_shape[1]*data_shape[2]*data_shape[3], data_shape[4])
	elif (len(data_shape) == 4):
		data = data.reshape(data_shape[0]*data_shape[1]*data_shape[2], data_shape[3])
		predictions = predictions.reshape(data_shape[0]*data_shape[1]*data_shape[2], data_shape[3])
	elif (len(data_shape) == 3):
		data = data.reshape(data_shape[0]*data_shape[1], data_shape[2])
		predictions = predictions.reshape(data_shape[0]*data_shape[1], data_shape[2])
	# These predictions are not persisted. Only used for metrics.
	split_metrics['r2'] = metrics.r2_score(data, predictions)
	split_metrics['mse'] = metrics.mean_squared_error(data, predictions)
	split_metrics['explained_variance'] = metrics.explained_variance_score(data, predictions)
	return split_metrics


def split_classification_plots(labels_processed, predictions, probabilities, analysis_type):
	predictions = predictions.flatten()
	probabilities = probabilities.flatten()
	split_plot_data = {}
	
	if (analysis_type == "classification_binary"):
		labels_processed = labels_processed.flatten()
		split_plot_data['confusion_matrix'] = metrics.confusion_matrix(labels_processed, predictions)
		fpr, tpr, _ = metrics.roc_curve(labels_processed, probabilities)
		precision, recall, _ = metrics.precision_recall_curve(labels_processed, probabilities)
	
	elif (analysis_type == "classification_multi"):
		# Flatten OHE labels for use with probabilities.
		labels_flat = labels_processed.flatten()
		fpr, tpr, _ = metrics.roc_curve(labels_flat, probabilities)
		precision, recall, _ = metrics.precision_recall_curve(labels_flat, probabilities)

		# Then convert unflat OHE to ordinal format for use with predictions.
		labels_ordinal = np.argmax(labels_processed, axis=1)
		split_plot_data['confusion_matrix'] = metrics.confusion_matrix(labels_ordinal, predictions)

	split_plot_data['roc_curve'] = {}
	split_plot_data['roc_curve']['fpr'] = fpr
	split_plot_data['roc_curve']['tpr'] = tpr
	split_plot_data['precision_recall_curve'] = {}
	split_plot_data['precision_recall_curve']['precision'] = precision
	split_plot_data['precision_recall_curve']['recall'] = recall
	return split_plot_data



import setuptools

# Makes it easy for contributors to install user-facing dependencies.
reqs = []
with open('requirements.txt') as f:
    for line in f:
        if not line.strip().startswith('#'):
            line = line.rstrip('\n')
            reqs.append(line)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aiqc",
    version="5.3.1",
    author="Layne Sadler",
    author_email="layne.sadler@gmail.com",
    description="End-to-end machine learning on your desktop or server.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://aiqc.readthedocs.io/",
    packages=setuptools.find_packages(),
    include_package_data=True,# Triggers `MANIFEST.in` file.
    python_requires='>=3.5, <=3.8.12', # (tf req Py3.5-3.8)
    license='BSD 3-Clause',
    # Version operands https://pip.pypa.io/en/stable/reference/pip_install/#requirement-specifiers
    # According to Python slack wizards, despite wheel-related warnings when installing aiqc on 
    # a fresh python env, I don't need to require users to install 'wheel'.
    install_requires=reqs,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Natural Language :: English",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 1 - Planning",
        "Framework :: Jupyter",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"
    ],
)



"""Functions that help with data ingestion, preprocessing, and inference."""
from os import path, listdir
from fsspec import filesystem
from natsort import natsorted
import pandas as pd
from textwrap import dedent
import numpy as np



def listify(supposed_lst:object=None):
	"""When only providing a single element, it's easy to forget to put it inside a list!"""
	if (supposed_lst is not None):
		if (not isinstance(supposed_lst, list)):
			supposed_lst = [supposed_lst]
		# If it was already a list, check it for emptiness and `None`.
		elif (isinstance(supposed_lst, list)):
			if (not supposed_lst):
				raise Exception("Yikes - The list you provided is empty.")
			if (None in supposed_lst):
				raise Exception(f"\nYikes - The list you provided contained `None` as an element.\n{supposed_lst}\n")
	# Allow entire list `is None` to pass through because we need it to trigger null conditions.
	return supposed_lst
    

def if_1d_make_2d(array:object):
	if (len(array.shape) == 1):
		array = array.reshape(array.shape[0], 1)
	return array


def sorted_file_list(dir_path:str):
	if (not path.exists(dir_path)):
		raise Exception(f"\nYikes - The path you provided does not exist according to `path.exists(dir_path)`:\n{dir_path}\n")
	file_path = path.abspath(dir_path)
	if (path.isdir(file_path) == False):
		raise Exception(f"\nYikes - The path that you provided is not a directory:{file_path}\n")
	file_paths = listdir(file_path)
	# prune hidden files and directories.
	file_paths = [f for f in file_paths if not f.startswith('.')]
	file_paths = [f for f in file_paths if not path.isdir(f)]
	if not file_paths:
		raise Exception(f"\nYikes - The directory that you provided has no files in it:{file_path}\n")
	# folder path is already absolute
	file_paths = [path.join(file_path, f) for f in file_paths]
	file_paths = natsorted(file_paths)
	return file_paths


def arr_validate(ndarray):
	if (type(ndarray).__name__ != 'ndarray'):
		raise Exception("\nYikes - The `ndarray` you provided is not of the type 'ndarray'.\n")
	if (ndarray.dtype.names is not None):
		raise Exception(dedent("""
		Yikes - Sorry, we do not support NumPy Structured Arrays.
		However, you can use the `dtype` dict and `column_names` to handle each column specifically.
		"""))
	if (ndarray.size == 0):
		raise Exception("\nYikes - The ndarray you provided is empty: `ndarray.size == 0`.\n")


def colIndices_from_colNames(column_names:list, desired_cols:list):
	desired_cols = listify(desired_cols)
	col_indices = [column_names.index(c) for c in desired_cols]
	return col_indices


def cols_by_indices(arr:object, col_indices:list):
	# Input and output 2D array. Fetches a subset of columns using their indices.
	# In the future if this needs to be adjusted to handle 3D array `[:,col_indices,:]`.
	subset_arr = arr[:,col_indices]
	return subset_arr


def pandas_stringify_columns(df, columns):
	"""
	- `columns` is user-defined.
	- Pandas will assign a range of int-based columns if there are no column names.
		So I want to coerce them to strings because I don't want both string and int-based 
		column names for when calling columns programmatically, 
		and more importantly, 'Exception: parquet must have string column names'
	"""
	cols_raw = df.columns.to_list()
	if (columns is None):
		# in case the columns were a range of ints.
		cols_str = [str(c) for c in cols_raw]
	else:
		cols_str = columns
	# dict from 2 lists
	cols_dct = dict(zip(cols_raw, cols_str))
	
	df = df.rename(columns=cols_dct)
	columns = df.columns.to_list()
	return df, columns


def df_validate(dataframe:object, column_names:list):
	if (dataframe.empty):
		raise Exception("\nYikes - The dataframe you provided is empty according to `df.empty`\n")

	if (column_names is not None):
		col_count = len(column_names)
		structure_col_count = dataframe.shape[1]
		if (col_count != structure_col_count):
			raise Exception(dedent(f"""
			Yikes - The dataframe you provided has <{structure_col_count}> columns,
			but you provided <{col_count}> columns.
			"""))


def df_set_metadata(dataframe:object, column_names:list=None, dtype:object=None):
	shape = {}
	shape['rows'], shape['columns'] = dataframe.shape[0], dataframe.shape[1]
	"""
	- Passes in user-defined columns in case they are specified.
	- Pandas auto-assigns int-based columns return a range when `df.columns`, 
		but this forces each column name to be its own str.
		"""
	dataframe, columns = pandas_stringify_columns(df=dataframe, columns=column_names)
	"""
	- At this point, user-provided `dtype` can be either a dict or a singular string/ class.
	- If columns are renamed, then dtype must used the renamed column names.
	- But a Pandas dataframe in-memory only has `dtypes` dict not a singular `dtype` str.
	- So we will ensure that there is 1 dtype per column.
	"""
	if (dtype is not None):
		# Accepts dict{'column_name':'dtype_str'} or a single str.
		try:
			dataframe = dataframe.astype(dtype)
		except:
			print("\nYikes - Failed to apply the dtypes you specified to the data you provided.\n")
			raise
		"""
		Check if any user-provided dtype against actual dataframe dtypes to see if conversions failed.
		Pandas dtype seems robust in comparing dtypes: 
		Even things like `'double' == dataframe['col_name'].dtype` will pass when `.dtype==np.float64`.
		Despite looking complex, category dtype converts to simple 'category' string.
		"""
		if (not isinstance(dtype, dict)):
			# Inspect each column:dtype pair and check to see if it is the same as the user-provided dtype.
			actual_dtypes = dataframe.dtypes.to_dict()
			for col_name, typ in actual_dtypes.items():
				if (typ != dtype):
					raise Exception(dedent(f"""
					Yikes - You specified `dtype={dtype},
					but Pandas did not convert it: `dataframe['{col_name}'].dtype == {typ}`.
					You can either use a different dtype, or try to set your dtypes prior to ingestion in Pandas.
					"""))
		elif (isinstance(dtype, dict)):
			for col_name, typ in dtype.items():
				if (typ != dataframe[col_name].dtype):
					raise Exception(dedent(f"""
					Yikes - You specified `dataframe['{col_name}']:dtype('{typ}'),
					but Pandas did not convert it: `dataframe['{col_name}'].dtype == {dataframe[col_name].dtype}`.
					You can either use a different dtype, or try to set your dtypes prior to ingestion in Pandas.
					"""))
	"""
	Testing outlandish dtypes:
	- `DataFrame.to_parquet(engine='auto')` fails on:
		'complex', 'longfloat', 'float128'.
	- `DataFrame.to_parquet(engine='auto')` succeeds on:
		'string', np.uint8, np.double, 'bool'.
	
	- But the new 'string' dtype is not a numpy type!
		so operations like `np.issubdtype` and `StringArray.unique().tolist()` fail.
	"""
	excluded_types = ['string', 'complex', 'longfloat', 'float128']
	actual_dtypes = dataframe.dtypes.to_dict().items()

	for col_name, typ in actual_dtypes:
		for et in excluded_types:
			if (et in str(typ)):
				raise Exception(dedent(f"""
				Yikes - You specified `dtype['{col_name}']:'{typ}',
				but aiqc does not support the following dtypes: {excluded_types}
				"""))
	"""
	Now, we take the all of the resulting dataframe dtypes and save them.
	Regardless of whether or not they were user-provided.
	Convert the classed `dtype('float64')` to a string so we can use it in `.to_pandas()`
	"""
	dtype = {k: str(v) for k, v in actual_dtypes}
	
	# Each object has the potential to be transformed so each object must be returned.
	return dataframe, columns, shape, dtype


def df_to_compressed_parquet_bytes(dataframe:object):
	"""
	- The Parquet file format naturally preserves pandas/numpy dtypes.
		Originally, we were using the `pyarrow` engine, but it has poor timedelta dtype support.
		https://towardsdatascience.com/stop-persisting-pandas-data-frames-in-csvs-f369a6440af5
	
	- Although `fastparquet` engine preserves timedelta dtype, but it does not work with BytesIO.
		https://github.com/dask/fastparquet/issues/586#issuecomment-861634507
	"""
	fs = filesystem("memory")
	temp_path = "memory://temp.parq"
	dataframe.to_parquet(temp_path, engine="fastparquet", compression="gzip", index=False)
	blob = fs.cat(temp_path)
	fs.delete(temp_path)
	return blob


def path_to_df(
	file_path:str
	, source_file_format:str
	, column_names:list
	, skip_header_rows:object
):
	"""
	Previously, I was using pyarrow for all tabular/ sequence file formats. 
	However, it had worse support for missing column names and header skipping.
	So I switched to pandas for handling csv/tsv, but read_parquet()
	doesn't let you change column names easily, so using pyarrow for parquet.
	""" 
	if (not path.exists(file_path)):
		raise Exception(f"\nYikes - The path you provided does not exist according to `path.exists(path)`:\n{file_path}\n")

	if (not path.isfile(file_path)):
		raise Exception(f"\nYikes - The path you provided is not a file according to `path.isfile(path)`:\n{file_path}\n")

	if (source_file_format == 'tsv') or (source_file_format == 'csv'):
		if (source_file_format == 'tsv') or (source_file_format is None):
			sep='\t'
			source_file_format = 'tsv' # Null condition.
		elif (source_file_format == 'csv'):
			sep=','

		df = pd.read_csv(
			filepath_or_buffer = file_path
			, sep = sep
			, names = column_names
			, header = skip_header_rows
		)
	elif (source_file_format == 'parquet'):
		if (skip_header_rows != 'infer'):
			raise Exception(dedent("""
			Yikes - The argument `skip_header_rows` is not supported for `source_file_format='parquet'`
			because Parquet stores column names as metadata.\n
			"""))
		df = pd.read_parquet(path=file_path, engine='fastparquet')
		df, columns = pandas_stringify_columns(df=df, columns=column_names)
	return df


def size_shift_defined(size_window:int=None, size_shift:int=None):
	"""Used by high level API classes."""
	if (
		((size_window is None) and (size_shift is not None))
		or 
		((size_window is not None) and (size_shift is None))
	):
		raise Exception("\nYikes - `size_window` and `size_shift` must be used together or not at all.\n")


def values_to_bins(array_to_bin:object, bin_count:int):
	"""
	Overwites continuous Label values with bin numbers for statification & folding.
	Switched to `pd.qcut` because `np.digitize` never had enough samples in the up the leftmost/right bin.
	"""
	# Make 1D for qcut.
	array_to_bin = array_to_bin.flatten()
	# For really unbalanced labels, I ran into errors where bin boundaries would be duplicates all the way down to 2 bins.
	# Setting `duplicates='drop'` to address this.
	bin_numbers = pd.qcut(x=array_to_bin, q=bin_count, labels=False, duplicates='drop')
	# When the entire `array_to_bin` is the same, qcut returns all nans!
	if (np.isnan(bin_numbers).any()):
		bin_numbers = None
	else:
		# Convert 1D array back to 2D for the rest of the program.
		bin_numbers = np.reshape(bin_numbers, (-1, 1))
	return bin_numbers


def stratifier_by_dtype_binCount(stratify_dtype:object, stratify_arr:object, bin_count:int=None):
	# Based on the dtype and bin_count determine how to stratify.
	# Automatically bin floats.
	if np.issubdtype(stratify_dtype, np.floating):
		if (bin_count is None):
			bin_count = 3
		stratifier = values_to_bins(array_to_bin=stratify_arr, bin_count=bin_count)
	# Allow ints to pass either binned or unbinned.
	elif (
		(np.issubdtype(stratify_dtype, np.signedinteger))
		or
		(np.issubdtype(stratify_dtype, np.unsignedinteger))
	):
		if (bin_count is not None):
			stratifier = values_to_bins(array_to_bin=stratify_arr, bin_count=bin_count)
		elif (bin_count is None):
			# Assumes the int is for classification.
			stratifier = stratify_arr
	# Reject binned objs.
	elif (np.issubdtype(stratify_dtype, np.number) == False):
		if (bin_count is not None):
			raise Exception(dedent("""
				Yikes - Your Label is not numeric (neither `np.floating`, `np.signedinteger`, `np.unsignedinteger`).
				Therefore, you cannot provide a value for `bin_count`.
			\n"""))
		elif (bin_count is None):
			stratifier = stratify_arr
	return stratifier, bin_count
	

def floats_only(label:object):
	# Prevent integer dtypes. It will ignore.
	label_dtypes = set(label.get_dtypes().values())
	for typ in label_dtypes:
		if (not np.issubdtype(typ, np.floating)):
			raise Exception(f"\nYikes - Interpolate can only be ran on float dtypes. Your dtype: <{typ}>\n")


def verify_attributes(interpolate_kwargs:dict):
	if (interpolate_kwargs['method'] == 'polynomial'):
		raise Exception("\nYikes - `method=polynomial` is prevented due to bug <https://stackoverflow.com/questions/67222606/interpolate-polynomial-forward-and-backward-missing-nans>.\n")
	if ((interpolate_kwargs['axis'] != 0) and (interpolate_kwargs['axis'] != 'index')):
		# This makes it so that you can run on sparse indices.
		raise Exception("\nYikes - `axis` must be either 0 or 'index'.\n")


def run_interpolate(dataframe:object, interpolate_kwargs:dict):
	# Interpolation does not require indices to be in order.
	dataframe = dataframe.interpolate(**interpolate_kwargs)
	return dataframe


def stage_data(
	splitset:object
	, job:object
	, samples:dict
	, library:str
	, key_train:str
):
	"""
	- Remember, you `.fit()` on either training data or all data (categoricals).
	- Then you transform the entire dataset because downstream processes may need the entire dataset:
		e.g. fit imputer to training data, then impute entire dataset so that categorical encoders can fit on entire dataset.
	- So we transform the entire dataset, then divide it into splits/ folds.
	- Then we convert the arrays to pytorch tensors if necessary. Subsetting with a list of indeces and `shape`
		work the same in both numpy and torch.
	"""
	# Labels - fetch and encode.
	if (splitset.supervision == "supervised"):
		label = splitset.label
		arr_labels = label.preprocess(
			samples = samples
			, _samples_train = samples[key_train]
			, _library = library
			, _job = job
		)

	# Features - fetch and encode.
	featureset = splitset.get_features()
	feature_count = len(featureset)
	features = []# expecting diff array shapes inside so it has to be list, not array.
	
	for feature in featureset:
		if (splitset.supervision == 'supervised'):
			arr_features = feature.preprocess(
				supervision = 'supervised'
				, samples = samples
				, _job = job
				, _samples_train = samples[key_train]
				, _library = library
			)	
		elif (splitset.supervision == 'unsupervised'):
			arr_features, arr_labels = feature.preprocess(
				supervision = 'unsupervised'
				, samples = samples
				, _job = job
				, _samples_train = samples[key_train]
				, _library = library
			)
		features.append(arr_features)
		# `arr_labels` is not appended because unsupervised analysis only supports 1 unsupervised feature.
	
	"""
	- Stage preprocessed data to be passed into the remaining Job steps.
	- Example samples dict entry: samples['train']['labels']
	- For each entry in the dict, fetch the rows from the encoded data.
	- Keras multi-input models accept input as a list. Not using nested dict for multiple
		features because it would be hard to figure out feature.id-based keys on the fly.
	""" 
	for split, indices in samples.items():
		if (feature_count == 1):
			samples[split] = {"features": arr_features[indices]}
		elif (feature_count > 1):
			# List of arrays is the preferred format for `tf.model.fit(x)` with multiple features.
			samples[split] = {"features": [arr_features[indices] for arr_features in features]}
		samples[split]['labels'] = arr_labels[indices]
	"""
	- Input shapes can only be determined after encoding has taken place.
	- `[0]` accessess the first sample in each array.
	- This shape does not influence the training loop's `batch_size`.
	- Shapes are used later by `get_model()` to initialize it.
	- Here the count refers to multimodal Features, not the number of columns.
	"""
	if (feature_count == 1):
		features_shape = samples[key_train]['features'][0].shape
	elif (feature_count > 1):
		features_shape = [arr_features[0].shape for arr_features in samples[key_train]['features']]
	input_shapes = {"features_shape": features_shape}

	label_shape = samples[key_train]['labels'][0].shape
	input_shapes["label_shape"] = label_shape

	return samples, input_shapes


def tabular_schemas_match(set_original, set_new):
	# Set can be either Label or Feature. Needs `columns` and `.get_dtypes`.
	cols_old = set_original.columns
	cols_new = set_new.columns
	if (cols_new != cols_old):
		raise Exception("\nYikes - New columns do not match original columns.\n")

	typs_old = set_original.get_dtypes()
	typs_new = set_new.get_dtypes()
	if (typs_new != typs_old):
		raise Exception(dedent("""
			Yikes - New dtypes do not match original dtypes.
			The Low-Level API methods for Dataset creation accept a `dtype` argument to fix this.
		"""))


def schemaNew_matches_schemaOld(splitset_new:object, splitset_old:object):
	# Get the new and old featuresets. Loop over them by index.
	features_new = splitset_new.get_features()
	features_old = splitset_old.get_features()

	if (len(features_new) != len(features_old)):
		raise Exception("\nYikes - Your new and old Splitsets do not contain the same number of Features.\n")

	for i, feature_new in enumerate(features_new):
		feature_old = features_old[i]
		feature_old_typ = feature_old.dataset.dataset_type
		feature_new_typ = feature_new.dataset.dataset_type
		if (feature_old_typ != feature_new_typ):
			raise Exception(f"\nYikes - New Feature dataset_type={feature_new_typ} != old Feature dataset_type={feature_old_typ}.\n")
		tabular_schemas_match(feature_old, feature_new)

		if (
			((len(feature_old.windows)>0) and (len(feature_new.windows)==0))
			or
			((len(feature_new.windows)>0) and (len(feature_old.windows)==0))
		):
			raise Exception("\nYikes - Either both or neither of Splitsets can have Windows attached to their Features.\n")

		if (((len(feature_old.windows)>0) and (len(feature_new.windows)>0))):
			window_old = feature_old.windows[-1]
			window_new = feature_new.windows[-1]
			if (
				(window_old.size_window != window_new.size_window)
				or
				(window_old.size_shift != window_new.size_shift)
			):
				raise Exception("\nYikes - New Window and old Window schemas do not match.\n")

	# Only verify Labels if the inference new Splitset provides Labels.
	# Otherwise, it may be conducting pure inference.
	label = splitset_new.label
	if (label is not None):
		label_new = label
		label_new_typ = label_new.dataset.dataset_type

		if (splitset_old.supervision == 'unsupervised'):
			raise Exception("\nYikes - New Splitset has Labels, but old Splitset does not have Labels.\n")

		elif (splitset_old.supervision == 'supervised'):
			label_old =  splitset_old.label
			label_old_typ = label_old.dataset.dataset_type
		
		if (label_old_typ != label_new_typ):
			raise Exception("\nYikes - New Label and original Label come from different `dataset_types`.\n")
		if (label_new_typ == 'tabular'):
			tabular_schemas_match(label_old, label_new)
